
//
//  HSMacro.h
//  HSCore
//
//  Created by Kyle Hankinson on 2018-10-05.
//  Copyright © 2018 Hankinsoft Development, Inc. All rights reserved.
//

#ifndef HSMacro_h
#define HSMacro_h

#import "Defer.h"
#import "MachTimer.h"
#import <objc/runtime.h>

#if !DEBUG
#define HSSAFESLEEP(statement) {error(statement)}
#else
#define HSSAFESLEEP(a)        [NSThread sleepForTimeInterval: (a)]
#endif

#define HSEmptyStringOnNil(a)   (a == nil ? @"" : a)

#define mustOverride() {\
NSString * logString = [NSString stringWithFormat:@"%@:%@ must be overridden in a subclass/category", NSStringFromClass(self.class), NSStringFromSelector(_cmd)];\
NSLog(@"%@", logString);\
@throw [NSException exceptionWithName:NSInvalidArgumentException reason: logString userInfo:nil]; \
}

#define LogEnterExitMeasure() {\
MachTimer * hsMeasureFunctionTimer = [MachTimer startTimer];\
[HSLogHelper.sharedInstance logMessage: @"-- <Enter %@:%@:%d>", NSStringFromClass([self class]), NSStringFromSelector(_cmd), __LINE__];\
pspdf_defer {\
NSLog(@"-- <Exit %@:%@:%d> took %0.2f seconds%@.", NSStringFromClass([self class]), NSStringFromSelector(_cmd), __LINE__, hsMeasureFunctionTimer.elapsedSeconds, hsMeasureFunctionTimer.elapsedSeconds > 1 ? @"(slow)" : @"");\
};\
}

#define LogIfSlowerThan(a) {\
MachTimer * hsMeasureFunctionTimer = [MachTimer startTimer];\
pspdf_defer {\
if(hsMeasureFunctionTimer.elapsedSeconds > (a)) {\
NSLog(@"-- <SLOW %@:%@:%d> took %0.2f seconds.", NSStringFromClass([self class]), NSStringFromSelector(_cmd), __LINE__, hsMeasureFunctionTimer.elapsedSeconds);\
}};\
}

// Compare
#define HSIsEqual(left,right) ((!left && !right) || (left == right) || [left isEqual: right])
#define HSIsDataEqual(left,right) ((!left && !right) || [left isEqualToData: right])
#define HSStringOrEmptyOnNil(a) (nil == a ? @"" : a)

#if TARGET_OS_IPHONE
#define DeviceIsIpad()          (UI_USER_INTERFACE_IDIOM() == UIUserInterfaceIdiomPad)
#else
#define DeviceIsIpad()          (FALSE)
#endif

// Supress performSelector-leaks warning -- based on
// https://stackoverflow.com/a/7933931/127853
#define SuppressPerformSelectorLeakWarning(Stuff) \
do { \
    _Pragma("clang diagnostic push") \
    _Pragma("clang diagnostic ignored \"-Warc-performSelector-leaks\"") \
    Stuff; \
    _Pragma("clang diagnostic pop") \
} while (0)

static inline void hs_executeOnceDaily(NSString * __nonnull dailyUniqueIdentifier, void (^ __nonnull block)(void))
{
#define kOneDay      86400       // 86400 = 24 hours ignoring daylight savings, etc, etc.
    NSUInteger lastRun = (NSUInteger)[NSUserDefaults.standardUserDefaults stringForKey: dailyUniqueIdentifier].integerValue;
    NSUInteger canRun = NSDate.date.timeIntervalSince1970 - kOneDay;

    if(lastRun < canRun)
    {
        block();

        NSString * timestampString = [NSString stringWithFormat: @"%ld", (unsigned long) NSDate.date.timeIntervalSince1970];

        [NSUserDefaults.standardUserDefaults setObject: timestampString
                                                forKey: dailyUniqueIdentifier];

        return;
    } // End of we have never run
} // End of hs_executeOnceDaily

static inline bool HSIsStringEqual(NSString * __nullable left, NSString * __nullable right)
{
    return ((!left && !right) || (left.length == right.length && [left isEqualToString: right]));
} // End of HSIsStringEqual

static inline bool HSIsStringEqualCaseInsensitive(NSString * __nullable left, NSString * __nullable right)
{
    return ((!left && !right) || (left.length == right.length && NSOrderedSame == [left caseInsensitiveCompare: right]));
} // End of HSIsStringEqualCaseInsensitive

static inline bool HSStringArrayContains(NSArray<NSString *>* __nullable array, NSString * __nullable checkFor, BOOL caseInsensitive)
{
    if(nil == array || 0 == array.count || 0 == checkFor.length)
    {
        return NO;
    } // End of array

    NSUInteger foundIndex = [array indexOfObjectPassingTest: ^BOOL(NSString * _Nonnull obj, NSUInteger idx, BOOL * _Nonnull stop) {
        if(caseInsensitive)
        {
            return HSIsStringEqualCaseInsensitive(obj, checkFor);
        }
        
        return HSIsStringEqual(obj, checkFor);
    }];

    return NSNotFound != foundIndex;
} // End of HSIsStringEqual

static inline bool HSIsStringArrayEqual(NSArray<NSString *>* __nullable left, NSArray<NSString *>* __nullable right, BOOL caseInsensitive)
{
    if(!left && !right)
    {
        return YES;
    } // End of is equal

    if(left.count != right.count)
    {
        return NO;
    }

    for(NSUInteger index = 0; index < left.count; ++index)
    {
        NSString * string1 = left[index];
        NSString * string2 = right[index];

        if(caseInsensitive)
        {
            if(!HSIsStringEqualCaseInsensitive(string1, string2))
            {
                return NO;
            }
        }
        else if(!HSIsStringEqual(string1, string2))
        {
            return NO;
        }
    }

    return YES;
} // End of HSIsStringArrayEqual

static inline bool HSValidRangeForString(NSRange range, NSString * __nullable string)
{
    if(NULL == string)
    {
        return NO;
    } // End of no string

    if(NSNotFound == range.location)
    {
        return NO;
    }

    if(range.length > string.length)
    {
        return NO;
    } // End of length is larger than the string

    if(NSMaxRange(range) > string.length)
    {
        return NO;
    }
    
    return YES;
} // End of HSValidRangeForString

static inline NSArray<Class> * __nullable ClassGetSubclasses(__nonnull Class parentClass)
{
    int numClasses = objc_getClassList(NULL, 0);
    Class *classes = NULL;
    
    classes = (Class*) malloc(sizeof(Class) * numClasses);
    
    numClasses = objc_getClassList(classes, numClasses);
    
    NSMutableArray *result = [NSMutableArray array];
    for (NSInteger i = 0; i < numClasses; i++) {
        Class superClass = classes[i];
        do{
            superClass = class_getSuperclass(superClass);
        } while(superClass && superClass != parentClass);
        
        if (superClass == nil) {
            continue;
        }
        [result addObject:classes[i]];
    }
    free(classes);
    return result;
}

static inline BOOL HSTestForClass(NSString * _Nonnull className)
{
    BOOL result = NO;
    Class testProbeClass = NSClassFromString(className);

    if(testProbeClass)
    {
        // Doing this little dance so we don't actually have to link
        // SenTestingKit in
        SEL selector = NSSelectorFromString(@"isTesting");
        NSMethodSignature *sig = [testProbeClass methodSignatureForSelector:selector];
        NSInvocation *invocation = [NSInvocation invocationWithMethodSignature:sig];
        [invocation setSelector:selector];
        [invocation invokeWithTarget:testProbeClass];

        [invocation getReturnValue: &result];
    }

    return result;
}

// Returns YES if we are currently being unittested.
// Adapted from https://stackoverflow.com/a/24688815/12257392
static inline BOOL HSAreWeBeingUnitTested()
{
  static BOOL answer = NO;

    static dispatch_once_t onceToken;
    dispatch_once(&onceToken, ^{
        if(HSTestForClass(@"XCTestProbe") || HSTestForClass(@"SenTestProbe"))
        {
            answer = YES;
        } // End of one of our test case classes exist
    });

  return answer;
}

#endif /* HSMacro_h */
